==> spacejunk-1.0.3.tar.gz: spacejunk-1.0.3/src/cmask.cpp <==
/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "cmask.h"

using namespace std;


Lexer & CMASK_parse(Lexer & p,CMask & m) {
    OUTLINE_Polygon * pol=NULL;
    Circle circ;
    PARSEbegin(Lexer,p);
    P1(OUTLINE_parsePolygon,&pol);
    P1(OUTLINE_parseCircle,&circ);
    m.push_back(make_pair(pol,circ));
    pol=NULL;
    MANYbegin;
    P1(OUTLINE_parsePolygon,&pol);
    P1(OUTLINE_parseCircle,&circ);
    m.push_back(make_pair(pol,circ));
    pol=NULL;
    MANYend;
    PARSEend;
    if (pol) delete pol;
    return p;
};

void CMASK_free(CMask * c) {
    for (CMask::iterator p=c->begin();p!=c->end();p++) delete p->first;
    delete c;
};


/*
 * Intersects a line and a circle centered in the origin.
 * @param r the circle radius.
 * @param p a point of the line.
 * @param dir vector indicating the direction of the line.
 * @param ti If ti!=NULL then after the call p+dir*(*ti) is one of the points of intersection.
 * @param tf If tf!=NULL then after the call p+dir*(*tf) is one of the points of intersection.
 * @returns true if dir!=0 the line intersects the circle. false otherwise.
 */
bool circle_line_inters(real r,const Vector2d & p,const Vector2d & dir,real * ti=NULL,real * tf=NULL) {
    // We have to solve the equation:
    //    0=dir.sqrDist()*t^2+2*dot(p,dir)*t+p.sqrDist()-r*r
    // First, we test if there are 2 real solutions
    // (2*dot(p,dir))^2-4*vel.sqrDist()*(p.sqrDist()-r*r))>0
    real b=2*dot(p,dir);
    real c=p.sqrDist()-r*r;
    real a=dir.sqrDist();
    real disc=b*b-4*a*c;
    if (disc<=0) return false;

    // Then we calculate the solutions
    // (-dot(p,dir)+-sqrt(disc))/(2*dir.sqrDist())
    real sqrtdisc=sqrt(disc);
    if (ti) *ti=(-b-sqrtdisc)/(2*a);
    if (tf) *tf=(-b+sqrtdisc)/(2*a);

    return true;
};


bool circle_swept_collision(real r1,real r2,const Vector2d & pos,const Vector2d & vel,real * t,Vector2d * normal) {
    // The time unit in these calculations is the time
    // passed between the moment when the bodies where at the
    // last test position and the moment when they reach their
    // current one. So we search a collision in the time interval
    // [0..1].

    // Test for intersection of gb2 path with the circumference of
    // radius gb1.hotspot.x+gb2.hotspot.x centered at origin.
    real r=r1+r2;

    if (vel==Vector2d()) {
        if (pos.sqrDist()<r*r) {
            if (normal) {
                *normal=pos;
                *normal/=normal->dist();
            }
            if (t) 
                *t=0;
            return true;
        } else 
            return false;
    }

    real ti,tf;
    // If trayectory segment intersects circle there is collision
    if (circle_line_inters(r,pos,vel,&ti,&tf) && 0<=tf && ti<=1) {
        if (normal) {
            *normal=pos+vel*ti;
            if (*normal!=Vector2d()) *normal/=normal->dist();
        }
        if (t) 
            *t=ti<0?0:ti;
        return true;
    } else 
        return false;
}


bool CMASK_intersect(const CMask * m1,const Matrix2x3 & m,real angleshift,const CMask * m2,const Vector2d & vel,real * t) {
    for (CMask::const_iterator i=m1->begin();i!=m1->end();i++) {
        for (CMask::const_iterator j=m2->begin();j!=m2->end();j++) {
            Vector2d pos=m*j->second.center-i->second.center;
			// if the enclosing disks collide test for the polygon intersection
            if (circle_swept_collision(i->second.r,j->second.r,pos,vel,t,NULL)) {
				// reverse one of the polygons, find the Minkowski sum and see
				// if the segment specified by the translation vector with an endpoint
				// at the origin intersects the result
                OUTLINE_PolygonReverser pr(m,j->first);
                OUTLINE_MKSum sum(i->first,-angleshift,&pr);
                real t0,t1;
                if (vel==0) {
                    if (sum.intersect_line(0,Vector2d(1,0),t0,t1) && t0<=0 && 0<=t1) {
                        if (t) 
                            *t=0;
                        return true;
                    }
                } else if (sum.intersect_line(0,vel,t0,t1) && t0<=1 && 0<=t1) {
                    if (t) 
                        *t=t0<0?0:t0;
                    return true;
                }
            };
        }
    }
    return false;
};


bool CMASK_intersect(const CMask * m,const Circle & c,const Vector2d & vel,real * t) {
    for (CMask::const_iterator i=m->begin();i!=m->end();i++) {
        if (circle_swept_collision(i->second.r,c.r,c.center-i->second.center,vel,t,NULL)) {
            if (vel!=0) {
                OUTLINE_Polygon pol(2);
                pol[0]=0;
                pol[1]=-vel;
                pol.compute();
                OUTLINE_PolygonTransformer pr(&pol);
                OUTLINE_MKSum sum(i->first,0,&pr);
                OUTLINE_MKSum::Index e;
                if (sum.closest_inside(c.center,e)) {
                    return true;
                } else {
                    Vector2d p0=sum[e];
                    Vector2d ve=sum[sum.next(e)]-p0;
                    real sin=dot(ve.rot90(),c.center-p0);
                    if (sin*sin/ve.sqrDist()<=c.r*c.r) return true;
                }
            } else {
                int e;
                if (i->first->closest_inside(c.center,e)) {
                    return true;
                } else {
                    Vector2d p0=(*i->first)[e];
                    Vector2d ve=(*i->first)[(e<i->first->size()-1?e+1:0)]-p0;
                    real sin=dot(ve.rot90(),c.center-p0);
                    if (sin*sin/ve.sqrDist()<=c.r*c.r) return true;
                }
            }
        }
    };
    return false;
}

Parser & CMaskLoader::parse(Parser & p,std::string * id) {
    string path,name;
    PARSEbegin(Parser,p);
    LS("<masks>");
    MANYbegin;
    TRY(S("<mask"))WS;
    S("name=")PATTTEXT(&name)S("path=")PATTTEXT(&path)S("/>");
    mask_paths[name]=path;
    MANYend;
    S("</masks>");
    PARSEend;
    return p;
}

CMask * CMaskLoader::create(const string & name) {
    string path=mask_paths[name];
    if (!path.length())
        return NULL;

    FILE * f=fopen(("data/"+path).c_str(),"r");
    if (!f) {
        CHERROR<<"Cannot open Mask file \""<<("data/"+path)<<"\": "<<SDL_GetError()<<ENDL;
        exit(1);
    }
    StreamTokenizer tok(f);
    Lexer p(&tok);
    CMask * m=new CMask();
    if (CMASK_parse(p,*m).error()) {
        fclose(f);
        CHERROR<<"ResourceManager error: Mask file "<<("data/"+path)<<" could not be loaded: "<<p.errorMessage()<<ENDL;
        exit(1);
    }
    fclose(f);

    return m;
};

==> spacejunk-1.0.5.tar.gz: spacejunk-1.0.5/src/cmask.cpp <==
/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "cmask.h"

using namespace std;


Lexer & CMASK_parse(Lexer & p,CMask & m) {
    OUTLINE_Polygon * pol=NULL;
    Circle circ;
    PARSEbegin(Lexer,p);
    P1(OUTLINE_parsePolygon,&pol);
    P1(OUTLINE_parseCircle,&circ);
    m.push_back(make_pair(pol,circ));
    pol=NULL;
    MANYbegin;
    P1(OUTLINE_parsePolygon,&pol);
    P1(OUTLINE_parseCircle,&circ);
    m.push_back(make_pair(pol,circ));
    pol=NULL;
    MANYend;
    PARSEend;
    if (pol) delete pol;
    return p;
};

void CMASK_free(CMask * c) {
    for (CMask::iterator p=c->begin();p!=c->end();p++) delete p->first;
    delete c;
};


/*
 * Intersects a line and a circle centered in the origin.
 * @param r the circle radius.
 * @param p a point of the line.
 * @param dir vector indicating the direction of the line.
 * @param ti If ti!=NULL then after the call p+dir*(*ti) is one of the points of intersection.
 * @param tf If tf!=NULL then after the call p+dir*(*tf) is one of the points of intersection.
 * @returns true if dir!=0 the line intersects the circle. false otherwise.
 */
bool circle_line_inters(real r,const Vector2d & p,const Vector2d & dir,real * ti=NULL,real * tf=NULL) {
    // We have to solve the equation:
    //    0=dir.sqrDist()*t^2+2*dot(p,dir)*t+p.sqrDist()-r*r
    // First, we test if there are 2 real solutions
    // (2*dot(p,dir))^2-4*vel.sqrDist()*(p.sqrDist()-r*r))>0
    real b=2*dot(p,dir);
    real c=p.sqrDist()-r*r;
    real a=dir.sqrDist();
    real disc=b*b-4*a*c;
    if (disc<=0) return false;

    // Then we calculate the solutions
    // (-dot(p,dir)+-sqrt(disc))/(2*dir.sqrDist())
    real sqrtdisc=sqrt(disc);
    if (ti) *ti=(-b-sqrtdisc)/(2*a);
    if (tf) *tf=(-b+sqrtdisc)/(2*a);

    return true;
};


bool circle_swept_collision(real r1,real r2,const Vector2d & pos,const Vector2d & vel,real * t,Vector2d * normal) {
    // The time unit in these calculations is the time
    // passed between the moment when the bodies where at the
    // last test position and the moment when they reach their
    // current one. So we search a collision in the time interval
    // [0..1].

    // Test for intersection of gb2 path with the circumference of
    // radius gb1.hotspot.x+gb2.hotspot.x centered at origin.
    real r=r1+r2;

    if (vel==Vector2d()) {
        if (pos.sqrDist()<r*r) {
            if (normal) {
                *normal=pos;
                *normal/=normal->dist();
            }
            if (t) 
                *t=0;
            return true;
        } else 
            return false;
    }

    real ti,tf;
    // If trayectory segment intersects circle there is collision
    if (circle_line_inters(r,pos,vel,&ti,&tf) && 0<=tf && ti<=1) {
        if (normal) {
            *normal=pos+vel*ti;
            if (*normal!=Vector2d()) *normal/=normal->dist();
        }
        if (t) 
            *t=ti<0?0:ti;
        return true;
    } else 
        return false;
}


bool CMASK_intersect(const CMask * m1,const Matrix2x3 & m,real angleshift,const CMask * m2,const Vector2d & vel,real * t) {
    for (CMask::const_iterator i=m1->begin();i!=m1->end();i++) {
        for (CMask::const_iterator j=m2->begin();j!=m2->end();j++) {
            Vector2d pos=m*j->second.center-i->second.center;
			// if the enclosing disks collide test for the polygon intersection
            if (circle_swept_collision(i->second.r,j->second.r,pos,vel,t,NULL)) {
				// reverse one of the polygons, find the Minkowski sum and see
				// if the segment specified by the translation vector with an endpoint
				// at the origin intersects the result
                OUTLINE_PolygonReverser pr(m,j->first);
                OUTLINE_MKSum sum(i->first,-angleshift,&pr);
                real t0,t1;
                if (vel==0) {
                    if (sum.intersect_line(0,Vector2d(1,0),t0,t1) && t0<=0 && 0<=t1) {
                        if (t) 
                            *t=0;
                        return true;
                    }
                } else if (sum.intersect_line(0,vel,t0,t1) && t0<=1 && 0<=t1) {
                    if (t) 
                        *t=t0<0?0:t0;
                    return true;
                }
            };
        }
    }
    return false;
};


bool CMASK_intersect(const CMask * m,const Circle & c,const Vector2d & vel,real * t) {
    for (CMask::const_iterator i=m->begin();i!=m->end();i++) {
        if (circle_swept_collision(i->second.r,c.r,c.center-i->second.center,vel,t,NULL)) {
            if (vel!=0) {
                OUTLINE_Polygon pol(2);
                pol[0]=0;
                pol[1]=-vel;
                pol.compute();
                OUTLINE_PolygonTransformer pr(&pol);
                OUTLINE_MKSum sum(i->first,0,&pr);
                OUTLINE_MKSum::Index e;
                if (sum.closest_inside(c.center,e)) {
                    return true;
                } else {
                    Vector2d p0=sum[e];
                    Vector2d ve=sum[sum.next(e)]-p0;
                    real sin=dot(ve.rot90(),c.center-p0);
                    if (sin*sin/ve.sqrDist()<=c.r*c.r) return true;
                }
            } else {
                int e;
                if (i->first->closest_inside(c.center,e)) {
                    return true;
                } else {
                    Vector2d p0=(*i->first)[e];
                    Vector2d ve=(*i->first)[(e<i->first->size()-1?e+1:0)]-p0;
                    real sin=dot(ve.rot90(),c.center-p0);
                    if (sin*sin/ve.sqrDist()<=c.r*c.r) return true;
                }
            }
        }
    };
    return false;
}

Parser & CMaskLoader::parse(Parser & p,std::string * id) {
    string path,name;
    PARSEbegin(Parser,p);
    LS("<masks>");
    MANYbegin;
    TRY(S("<mask"))WS;
    S("name=")PATTTEXT(&name)S("path=")PATTTEXT(&path)S("/>");
    mask_paths[name]=path;
    MANYend;
    S("</masks>");
    PARSEend;
    return p;
}

CMask * CMaskLoader::create(const string & name) {
    string path=mask_paths[name];
    if (!path.length())
        return NULL;

    FILE * f=fopen((GetDataPath()+path).c_str(),"r");
    if (!f) {
        CHERROR<<"Cannot open Mask file \""<<(GetDataPath()+path)<<"\": "<<SDL_GetError()<<ENDL;
        exit(1);
    }
    StreamTokenizer tok(f);
    Lexer p(&tok);
    CMask * m=new CMask();
    if (CMASK_parse(p,*m).error()) {
        fclose(f);
        CHERROR<<"ResourceManager error: Mask file "<<(GetDataPath()+path)<<" could not be loaded: "<<p.errorMessage()<<ENDL;
        exit(1);
    }
    fclose(f);

    return m;
};
